/*****************************************************************
 * FILE: alignalign_time.c
 *****************************************************************
 * Copyright 2005 by Dean Starrett, Travis Wheeler
 * University of Arizona
 * Department of Computer Science
 * Tucson, AZ  85721
 *****************************************************************
 * Terms of use:
 * This software may be used for non-commercial purposes.
 * Other use, redistribution, or modification may be done only 
 * with the author's prior approval. (contact twheeler@cs.arizona.edu)
 *
 * Released: 17 May 2007
 * Version:  1.0.0
 * 
 ****************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>       /* strlen, strcat  */
#include <ctype.h>        /* isspace         */
#include "alignalign.h"

/*
 * alignalign_time
 */
 
int alignalign_time (char ***alignment_ptr) 
{
  int      h, i, j;
  Shape  **D1 = NULL;                     /* dynamic prog table row 1 */
  Shape  **D2 = NULL;                     /* dynamic prog table row 2 */
  Shape  **Dtemp = NULL;                  /* to swap table rows 1 & 2 */
  Shape   *p = NULL;                      /* Shape list iterator      */
  Shape   *q = NULL;                      /* Shape list iterator      */
  int c = 0;                     /* temporary cost holder    */
  int current_c = 0;             /* temporary cost holder    */
  int      dom_prune = AA_FALSE;          /* dominance pruning flag   */
  int      bound_prune = AA_FALSE;        /* bound pruning flag       */
  Shape   *header;
  aa_ushort s[AA_MAXSEQNUM];         /* temporary variable  - TW : s is an array of block positions (a shape w/o all the extra info)     */
  int naive_upper_bound;      /* cost of the naive alignment */
  int optimistic_upper_bound;
  int pessimistic_upper_bound;
  int min = 0;
  Shape *recover1, *recover2, *recover3, *pmin;
  int alignlen;
  int       Aindex, Bindex, op;
  int max, aflag, bflag;

  /* compute the cost of naive re-alignment */
  naive_upper_bound = alignmentCost(aa_A, aa_B, aa_K, aa_L, aa_M, aa_N);

  pessimistic_upper_bound = 
    pessimistic_optimistic(aa_A, aa_B, aa_K, aa_L, aa_M, aa_N, 1.0);

  optimistic_upper_bound =    /* builds lower-bound tables, which are conserved for later use in bound pruning*/
    reverse_optimistic_table(aa_A, aa_B, aa_K, aa_L, aa_M, aa_N, &aa_optH, &aa_optV, &aa_optD);

  if (pessimistic_upper_bound < naive_upper_bound) {
    aa_cost_upper_bound = pessimistic_upper_bound;
  }
  else { aa_cost_upper_bound = naive_upper_bound; }
  if (optimistic_upper_bound < aa_cost_upper_bound) {
    aa_cost_upper_bound = optimistic_upper_bound;
  }

  /* allocate dynamic programming table */
  D1 = aa_malloc((aa_N+1) * sizeof(Shape*));
  D2 = aa_malloc((aa_N+1) * sizeof(Shape*));
  for (j = 0; j <= aa_N; j++) {
    D1[j] = NULL;
    D1[j] = nodemalloc();
    NEXT(D1[j]) = D1[j]; /* initially, entry is empty circular list */
    D2[j] = NULL;
    D2[j] = nodemalloc();
    NEXT(D2[j]) = D2[j]; /* initially, entry is empty circular list */
  }

  /* initialize entry (0,0) with the flush shape and cost 0 */
  p = nodemalloc();
  NEXT(D1[0]) = p; /* header node */
  NEXT(p) = D1[0]; /* circular loop with a single element */
  RECOVER(p) = NULL;
  C(p) = 0;
  for (i = 0; i < aa_K+aa_L; i++) { (S(p))[i] = 0; } /* flush shape */
  
//  printf ("table of size %d x %d\n-------------------\n", aa_M+1, aa_N+1);
//  printf ("upper bound is %d\n", aa_cost_upper_bound);
  /* fill in the dynamic programming table (row major order) */
  for (i = 0; i < aa_M; i++) {
    for (j = 0; j < aa_N; j++) {

      header = D1[j];

      if (NULL != header) {
		p = NEXT(D1[j]);
	
/*	    printf ("From %d, %d: \n", i, j);
     	int ii;
		for (ii=0; ii<aa_K; ii++) {
			printf ("%c,", aa_A[ii][i]);
		}
		printf (" vs. ");
		for (ii=0; ii<aa_L; ii++) {
			printf ("%c,", aa_B[ii][j]);
		}	
		printf ("\n");
*/	
		while (p != header) {//foreach shape

		  /* save cost of current shape */
		  current_c = C(p);
		  /* resuse cost field for reference count (number of shapes */
		  /* propagated from this shape).                             */
		  C(p) = 0;
		  
		  /*** propagate from D[i][j] to D[i+1][j] ***/
		  /* determine new shape */
		  newshape(aa_A, NULL, i, -1, S(p), s);
		  /* determine new cost */
		  c = current_c + subs(i,j,AA_VERT) + gaps(i,j,S(p),AA_VERT);
		  /* test for dominance prunning */
		  dom_prune = dominance(D2[j], s, c, i, j);
		  if (!dom_prune) {
		    bound_prune = bound(s, c, i+1, j);
		    if (!bound_prune) {
		      /* insert new shape into the list at D[i+1][j] */
		      insert(D2[j], s, c, p); // insert shape s into cell D2[j], with cost c, and parent p.
		      //if (i >= 220 && j+1 == 198) { 
//		      printf ("\tA From %d,%d to %d,%d,  cost %d (%d)\n", i, j, i+1, j, c, current_c); //}
//	          printf ("A From %d,%d to %d,%d,  cost %d (%d) [+ %d = %d - %d = %d  <? %d]\n", i, j, i+1, j, c, current_c, aa_optV[i+1][j], c+aa_optV[i+1][j], get_Vgap(s, c, i+1, j), c+aa_optV[i+1][j]-get_Vgap(s, c, i+1, j), aa_cost_upper_bound ); //}

		    }
		    else { bound_prune = AA_FALSE; }
		  }
	
		  /*** propagate from D[i][j] to D[i][j+1] ***/
		  /* determine new shape */
		  newshape(NULL, aa_B, -1, j, S(p), s);
		  /* determine new cost */
		  c = current_c + subs(i,j,AA_HORIZ) + gaps(i,j,S(p),AA_HORIZ);
		  /* test for dominance prunning */
		  dom_prune = dominance(D1[j+1], s, c, i, j);
		  if (!dom_prune) { 
		    bound_prune = bound(s, c, i, j+1);
		    if (!bound_prune) {
		      /* insert new shape into the list at D[i][j+1] */
		      insert(D1[j+1], s, c, p);

		      //if (i >= 220 && j+1 >= 198) { 
//		      printf ("\tB from %d,%d to %d,%d,  cost %d (%d) \n", i, j, i, j+1, c, current_c ); //}
//		      printf ("B from %d,%d to %d,%d,  cost %d (%d) [+ %d = %d - %d = %d <? %d]\n", i, j, i, j+1, c, current_c, aa_optH[i][j+1], c+aa_optH[i][j+1],get_Hgap(s, c, i, j+1), c+aa_optH[i][j+1]-get_Hgap(s, c, i, j+1), aa_cost_upper_bound ); //}
		    }
		  }
	
		  /*** propagate from D[i][j] to D[i+1][j+1] ***/
		  /* determine new shape */
		  newshape(aa_A, aa_B, i, j, S(p), s);
		  /* determine new cost */
		  c = current_c + subs(i,j,AA_DIAG) + gaps(i,j,S(p),AA_DIAG);
		  /* test for dominance prunning */
		  dom_prune = dominance(D2[j+1], s, c, i, j);
		  if (!dom_prune) {
		    bound_prune = bound(s, c, i+1, j+1);
		    if (!bound_prune) {
		      /* insert new shape into the list at D[i+1][j+1] */
		      insert(D2[j+1], s, c, p);
//   				if (i >= 220 && j+1 >= 198) { 
//				printf ("\tC From %d,%d to %d,%d,  cost %d (%d)\n", i, j, i+1, j+1, c, current_c); //}
// 		        printf ("C From %d,%d to %d,%d,  cost %d (%d) [+ %d = %d - %d = %d  <? %d]\n", i, j, i+1, j+1, c, current_c, aa_optD[i+1][j+1], c+aa_optD[i+1][j+1], get_Dgap(s, c, i+1, j+1), c+aa_optD[i+1][j+1]-get_Dgap(s, c, i+1, j+1), aa_cost_upper_bound ); //}
	
		    }
		  }
	
		  /* if nothing was propagated, reclaim the shape */
		  if (1 > C(p)) { q = NEXT(p); reclaim(p); p = q; }
		  else { p = NEXT(p); }
		}
      }
    } /* for (j = 0; j < n; j++) */

    /* j == n, so only propagate to D[i+1][j] */
    header = D1[j];

    if (NULL != header) {
      p = NEXT(D1[j]);

      /* loop through the shape list at D[i][j] */
      while (p != header) {
		/* save cost of current shape */
		current_c = C(p);
		/* reuse cost field for reference count (number of shapes */
		/* propagated from this shape).                             */
		C(p) = 0;
		
		/*** propagate from D[i][j] to D[i+1][j] ***/
		/* determine new shape */
		newshape(aa_A, NULL, i, -1, S(p), s);
		/* determine new cost */
		c = current_c + subs(i,j,AA_VERT) + gaps(i,j,S(p),AA_VERT);
		/* test for dominance prunning */
		dom_prune = dominance(D2[j], s, c, i, j);
		if (!dom_prune) { 
		  bound_prune = bound(s, c, i+1, j);
		  if (!bound_prune) {
		    /* insert new shape into the list at D[i+1][j] */
		    insert(D2[j], s, c, p);
   		    //if (i >= 220 && j >= 198) { 
//	        printf ("\tD From %d,%d to %d,%d,  cost %d (%d) \n", i, j, i+1, j, c, current_c); //}
//	        printf ("D From %d,%d to %d,%d,  cost %d (%d) [+ %d = %d - %d = %d  <? %d]\n", i, j, i+1, j, c, current_c, aa_optV[i+1][j], c+aa_optV[i+1][j], get_Vgap(s, c, i+1, j), c+aa_optV[i+1][j]-get_Vgap(s, c, i+1, j), aa_cost_upper_bound ); //}

		  }
		}
		
		/* if nothing was propagated, reclaim the shape */
		if (1 > C(p)) { q = NEXT(p); reclaim(p); p = q; }
		else { p = NEXT(p); }
      }
    }
    
    /* swap rows D1 & D2 */
    Dtemp = D1;
    D1 = D2;
    D2 = Dtemp;
    
    /* clean up D2 */
    for (h = 0; h <= aa_N; h++) { NEXT(D2[h]) = D2[h]; }


  } /* for (i = 0; i < m; i++) */


  /* i == m, so only propagate to D[i][j+1] */
  for (j = 0; j < aa_N; j++) {
    header = D1[j];

    if (NULL != header) {
      p = NEXT(D1[j]);
      
//	  printf ("From %d, %d: \n", i, j);         
      while (p != header) {
		/* save cost of current shape */
		current_c = C(p);
		/* reuse cost field for reference count (number of shapes */
		/* propagated from this shape).                             */
		C(p) = 0;
		
		/*** propagate from D[i][j] to D[i][j+1] ***/
		/* determine new shape */
		newshape(NULL, aa_B, -1, j, S(p), s);
		/* determine new cost */
		c = current_c + subs(i,j,AA_HORIZ) + gaps(i,j,S(p),AA_HORIZ);
		/* test for dominance prunning */
		dom_prune = dominance(D1[j+1], s, c, i, j);
		if (!dom_prune) { 
		  bound_prune = bound(s, c, i, j+1);
		  if (!bound_prune) {
		    /* insert new shape into the list at D[i][j+1] */
		    insert(D1[j+1], s, c, p);
 		    //if (i >= 220 && j+1 >= 198) {
//		    	 printf ("\tE From %d,%d to %d,%d,  cost %d (%d)\n", i, j, i, j+1, c, current_c);// }
//				 printf ("E from %d,%d to %d,%d,  cost %d (%d) [+ %d = %d - %d = %d  <? %d]\n", i, j, i, j+1, c, current_c, aa_optH[i][j+1], c+aa_optH[i][j+1],get_Hgap(s, c, i, j+1), c+aa_optH[i][j+1]-get_Hgap(s, c, i, j+1), aa_cost_upper_bound ); //}
	 		    	 
		  }
		}
		
		/* if nothing was propagated, reclaim the shape */
		if (1 > C(p)) { q = NEXT(p); reclaim(p); p = q; }
		else { p = NEXT(p); }
      }
    }
  } /* for (j = 0; j < n; j++) */

/*
fflush (stdout);
fflush (stderr);
*/
  /* get alignment length, reset recovery pointers for forward recovery */
  alignlen = 0;

  /* find the optimal solution shape in D[m][n] */
  header = D1[aa_N];
  min = C(NEXT(D1[aa_N]));
  pmin = p = NEXT(header);
  while (p != header) { 
    if (C(p) < min) { min = C(p); pmin = p; } 
    p = NEXT(p);
  }

  recover1 = NULL;
  recover2 = NULL;
  recover3 = pmin;
  if (NULL != recover3) {
    recover2 = RECOVER(recover3);
    RECOVER(recover3) = NULL;
  }
  else {
    fprintf(stderr, "alignment length is 0!  Exiting...\n");
    exit(1);
  }
  if (NULL != recover2) {
    recover1 = RECOVER(recover2);
    RECOVER(recover2) = recover3;
    alignlen++;
  }
  else { recover2 = recover3; }

  while (NULL != recover1) {
    recover3 = recover2;
    recover2 = recover1;
    recover1 = RECOVER(recover1);
    RECOVER(recover2) = recover3;
    alignlen++;
  }

  /* allocate alignment matrix (to store final result) */
  *alignment_ptr = aa_malloc((aa_K+aa_L) * sizeof(char *));
  for (i = 0; i < aa_K+aa_L; i++) {
    (*alignment_ptr)[i] = aa_malloc((alignlen + 1) * sizeof(char));
  }
  
  aa_script = aa_malloc( alignlen * sizeof(unsigned char));
  
  /* recover the solution */
  recover1 = RECOVER(recover2);
  j = Aindex = Bindex = 0;

  while (NULL != recover1) {
    
    /* determine operation (sub, ins, del) that produced this shape */
    //TW - figure out the max block, then if a's max_block = b's max_block = the overall max block, then Sub, else ...
    max = 0;
    aflag = bflag = AA_FALSE;
    for (i = 0; i < aa_K + aa_L; i++) {
      if (max < (S(recover1))[i]) { max = (S(recover1))[i]; }
    }
    for (i = 0; i < aa_K; i++) {
      if (max == (S(recover1))[i]) { aflag = AA_TRUE; }
    }
    for (i = aa_K; i < aa_K + aa_L; i++) {
      if (max == (S(recover1))[i]) { bflag = AA_TRUE; }
    }

    if (aflag && bflag) { op = AA_SUB; }
    else if (aflag)     { op = AA_DEL; }
    else if (bflag)     { op = AA_INS; }
    else {
      fprintf(stderr, "*** internal bug in alignment recovery *** %d\n\n", j);
      exit(1);
    }

	aa_script[j] = op;

    /* fill in the alignment according to the operation (sub, ins, del) */
   switch (op) {
	    case AA_INS:
	      for (i=0; i < aa_K; i++) { (*alignment_ptr)[i][j] = AA_EPS; }
	      for (i=0; i < aa_L; i++) { (*alignment_ptr)[i + aa_K][j] = aa_B[i][Bindex]; } 
	      Bindex++;
	      //fprintf(stderr, "-");
	      break;
	
	    case AA_DEL:
	      for (i=0; i < aa_K; i++) { (*alignment_ptr)[i][j] = aa_A[i][Aindex]; }
	      Aindex++;
	      for (i=0; i < aa_L; i++) { (*alignment_ptr)[i + aa_K][j] = AA_EPS; }
	      //fprintf(stderr, "|");
	      break;
	
	    case AA_SUB:
	      for (i=0; i < aa_K; i++) { (*alignment_ptr)[i][j] = aa_A[i][Aindex]; }
	      Aindex++;
	      for (i=0; i < aa_L; i++) { (*alignment_ptr)[i + aa_K][j] = aa_B[i][Bindex]; }
	      Bindex++;
	      //fprintf(stderr, "\\");
	      break;
    };

    j++;
    recover2 = recover1;
    recover1 = RECOVER(recover1);
  }
  //printf("\n");

  /*
   * add cleanup code
   */
  free(D1);
  free(D2);
    
  nodepoolfree(0);

  for (i = 0; i <= aa_M; i++) {
    aa_free(aa_optH[i], (aa_N+1) * sizeof(int));
    aa_free(aa_optV[i], (aa_N+1) * sizeof(int));
    aa_free(aa_optD[i], (aa_N+1) * sizeof(int));
  }

  aa_free(aa_optH, (aa_M+1) * sizeof(int *));
  aa_free(aa_optV, (aa_M+1) * sizeof(int *));
  aa_free(aa_optD, (aa_M+1) * sizeof(int *));


  return alignlen;
}
